"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tsutils_1 = require("tsutils");
const unsafe_assignment_rule_1 = require("./unsafe-assignment-rule");
const unsafeIndexAssignmentFunc = (indexKind, destinationType, sourceType, checker) => {
    const destinationIndexInfo = checker.getIndexInfoOfType(destinationType, indexKind);
    const sourceIndexInfo = checker.getIndexInfoOfType(sourceType, indexKind);
    // eslint-disable-next-line functional/no-conditional-statement
    if (destinationIndexInfo === undefined) {
        return false;
    }
    // eslint-disable-next-line functional/no-conditional-statement
    if (sourceIndexInfo === undefined) {
        return false;
    }
    const destinationTypeHasReadonlyIndexSignature = destinationIndexInfo.isReadonly;
    const sourceTypeHasReadonlyIndexSignature = sourceIndexInfo.isReadonly;
    return (!sourceTypeHasReadonlyIndexSignature &&
        destinationTypeHasReadonlyIndexSignature);
};
const unsafePropertyAssignmentFunc = (
// eslint-disable-next-line @typescript-eslint/ban-types
destinationProperty, 
// eslint-disable-next-line @typescript-eslint/ban-types
sourceProperty, destinationType, sourceType, checker) => {
    // eslint-disable-next-line functional/no-conditional-statement
    if (sourceProperty === undefined) {
        return false;
    }
    const destinationPropIsReadonly = (0, tsutils_1.isPropertyReadonlyInType)(destinationType, destinationProperty.getEscapedName(), checker);
    const sourcePropIsReadonly = (0, tsutils_1.isPropertyReadonlyInType)(sourceType, sourceProperty.getEscapedName(), checker);
    return !sourcePropIsReadonly && destinationPropIsReadonly;
};
/**
 * An ESLint rule to ban unsafe assignment from mutable to readonly types.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noUnsafeMutableReadonlyAssignment = {
    meta: {
        type: "problem",
        docs: {
            description: "Bans unsafe assignment from mutable to readonly types.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringCallExpression: "Passing a mutable type to a function that expects a readonly type can lead to unexpected mutation in the readonly value.",
            errorStringAssignmentExpression: "Assigning a mutable type to a readonly type can lead to unexpected mutation in the readonly value.",
            errorStringVariableDeclaration: "Using a mutable type to initialize a readonly type can lead to unexpected mutation in the readonly value.",
            errorStringArrowFunctionExpression: "Returning a mutable type from a function that returns a readonly type can lead to unexpected mutation in the readonly value.",
            errorStringTSAsExpression: "Asserting a mutable type to a readonly type can lead to unexpected mutation in the readonly value.",
            errorStringTSTypeAssertion: "Asserting a mutable type to a readonly type can lead to unexpected mutation in the readonly value.",
        },
        schema: [],
    },
    create: (0, unsafe_assignment_rule_1.createNoUnsafeAssignmentRule)(unsafePropertyAssignmentFunc, unsafeIndexAssignmentFunc),
};
exports.default = noUnsafeMutableReadonlyAssignment;
