"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tsutils_1 = require("tsutils");
const unsafe_assignment_rule_1 = require("./unsafe-assignment-rule");
const unsafeIndexAssignmentFunc = (indexKind, destinationType, sourceType, checker) => {
    const destinationIndexInfo = checker.getIndexInfoOfType(destinationType, indexKind);
    const destinationTypeHasReadonlyIndexSignature = destinationIndexInfo !== undefined
        ? destinationIndexInfo.isReadonly
        : false;
    const sourceIndexInfo = checker.getIndexInfoOfType(sourceType, indexKind);
    const sourceTypeHasReadonlyIndexSignature = sourceIndexInfo !== undefined ? sourceIndexInfo.isReadonly : false;
    return (sourceTypeHasReadonlyIndexSignature &&
        !destinationTypeHasReadonlyIndexSignature);
};
const unsafePropertyAssignmentFunc = (
// eslint-disable-next-line @typescript-eslint/ban-types
destinationProperty, 
// eslint-disable-next-line @typescript-eslint/ban-types
sourceProperty, destinationType, sourceType, checker) => {
    // eslint-disable-next-line functional/no-conditional-statement
    if (sourceProperty === undefined) {
        return false;
    }
    const destinationPropIsReadonly = (0, tsutils_1.isPropertyReadonlyInType)(destinationType, destinationProperty.getEscapedName(), checker);
    const sourcePropIsReadonly = (0, tsutils_1.isPropertyReadonlyInType)(sourceType, sourceProperty.getEscapedName(), checker);
    return sourcePropIsReadonly && !destinationPropIsReadonly;
};
/**
 * An ESLint rule to ban unsafe assignment from readonly to mutable types.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noUnsafeReadonlyMutableAssignment = {
    meta: {
        type: "problem",
        docs: {
            description: "Bans unsafe assignment from readonly to mutable types.",
            recommended: "error",
            url: "https://github.com/danielnixon/eslint-plugin-total-functions",
        },
        messages: {
            errorStringCallExpression: "Passing a readonly type to a function that expects a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringAssignmentExpression: "Assigning a readonly type to a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringVariableDeclaration: "Using a readonly type to initialize a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringArrowFunctionExpression: "Returning a readonly type from a function that returns a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringTSAsExpression: "Asserting a readonly type to a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringTSTypeAssertion: "Asserting a readonly type to a mutable type can lead to unexpected mutation in the readonly value.",
        },
        schema: [],
    },
    create: (0, unsafe_assignment_rule_1.createNoUnsafeAssignmentRule)(unsafePropertyAssignmentFunc, unsafeIndexAssignmentFunc),
};
exports.default = noUnsafeReadonlyMutableAssignment;
